use oiseau::cache::Cache;
use crate::model::moderation::AuditLogEntry;
use crate::model::{
    Error, Result, auth::User, permissions::FinePermission,
    communities_permissions::CommunityPermission, channels::Channel,
};
use crate::{auto_method, DataManager};

use oiseau::{PostgresRow, execute, get, query_row, query_rows, params};

impl DataManager {
    /// Get a [`Channel`] from an SQL row.
    pub(crate) fn get_channel_from_row(x: &PostgresRow) -> Channel {
        Channel {
            id: get!(x->0(i64)) as usize,
            community: get!(x->1(i64)) as usize,
            owner: get!(x->2(i64)) as usize,
            created: get!(x->3(i64)) as usize,
            minimum_role_read: get!(x->4(i32)) as u32,
            minimum_role_write: get!(x->5(i32)) as u32,
            position: get!(x->6(i32)) as usize,
            members: serde_json::from_str(&get!(x->7(String))).unwrap(),
            title: get!(x->8(String)),
            last_message: get!(x->9(i64)) as usize,
        }
    }

    auto_method!(get_channel_by_id(usize as i64)@get_channel_from_row -> "SELECT * FROM channels WHERE id = $1" --name="channel" --returns=Channel --cache-key-tmpl="atto.channel:{}");

    /// Get all member profiles from a channel members list.
    pub async fn fill_members(
        &self,
        members: &Vec<usize>,
        ignore_users: Vec<usize>,
    ) -> Result<Vec<User>> {
        let mut out = Vec::new();

        for member in members {
            if ignore_users.contains(member) {
                continue;
            }

            out.push(self.get_user_by_id(member.to_owned()).await?);
        }

        Ok(out)
    }

    /// Get all channels by community.
    ///
    /// # Arguments
    /// * `community` - the ID of the community to fetch channels for
    pub async fn get_channels_by_community(&self, community: usize) -> Result<Vec<Channel>> {
        let conn = match self.0.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = query_rows!(
            &conn,
            "SELECT * FROM channels WHERE community = $1 ORDER BY position ASC",
            &[&(community as i64)],
            |x| { Self::get_channel_from_row(x) }
        );

        if res.is_err() {
            return Err(Error::GeneralNotFound("channel".to_string()));
        }

        Ok(res.unwrap())
    }

    /// Get all channels by user.
    ///
    /// # Arguments
    /// * `user` - the ID of the user to fetch channels for
    pub async fn get_channels_by_user(&self, user: usize) -> Result<Vec<Channel>> {
        let conn = match self.0.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = query_rows!(
            &conn,
            "SELECT * FROM channels WHERE (owner = $1 OR members LIKE $2) AND community = 0 ORDER BY last_message DESC",
            params![&(user as i64), &format!("%{user}%")],
            |x| { Self::get_channel_from_row(x) }
        );

        if res.is_err() {
            return Err(Error::GeneralNotFound("channel".to_string()));
        }

        Ok(res.unwrap())
    }

    /// Get a channel given its `owner` and a member.
    ///
    /// # Arguments
    /// * `owner` - the ID of the owner
    /// * `member` - the ID of the member
    pub async fn get_channel_by_owner_member(
        &self,
        owner: usize,
        member: usize,
    ) -> Result<Channel> {
        let conn = match self.0.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = query_row!(
            &conn,
            "SELECT * FROM channels WHERE owner = $1 AND members = $2 AND community = 0 ORDER BY created DESC",
            params![&(owner as i64), &format!("[{member}]")],
            |x| { Ok(Self::get_channel_from_row(x)) }
        );

        if res.is_err() {
            return Err(Error::GeneralNotFound("channel".to_string()));
        }

        Ok(res.unwrap())
    }

    /// Create a new channel in the database.
    ///
    /// # Arguments
    /// * `data` - a mock [`Channel`] object to insert
    pub async fn create_channel(&self, data: Channel) -> Result<()> {
        let user = self.get_user_by_id(data.owner).await?;

        // check user permission in community
        if data.community != 0 {
            let membership = self
                .get_membership_by_owner_community(user.id, data.community)
                .await?;

            if !membership.role.check(CommunityPermission::MANAGE_CHANNELS)
                && !user.permissions.check(FinePermission::MANAGE_CHANNELS)
            {
                return Err(Error::NotAllowed);
            }
        }
        // check members
        else {
            for member in &data.members {
                if self
                    .get_userblock_by_initiator_receiver(member.to_owned(), data.owner)
                    .await
                    .is_ok()
                {
                    return Err(Error::NotAllowed);
                }
            }
        }

        // ...
        let conn = match self.0.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = execute!(
            &conn,
            "INSERT INTO channels VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10)",
            params![
                &(data.id as i64),
                &(data.community as i64),
                &(data.owner as i64),
                &(data.created as i64),
                &(data.minimum_role_read as i32),
                &(data.minimum_role_write as i32),
                &(data.position as i32),
                &serde_json::to_string(&data.members).unwrap(),
                &data.title,
                &(data.last_message as i64)
            ]
        );

        if let Err(e) = res {
            return Err(Error::DatabaseError(e.to_string()));
        }

        Ok(())
    }

    pub async fn delete_channel(&self, id: usize, user: &User) -> Result<()> {
        let channel = self.get_channel_by_id(id).await?;

        // check user permission in community
        if user.id != channel.owner {
            let membership = self
                .get_membership_by_owner_community(user.id, channel.community)
                .await?;

            if !membership.role.check(CommunityPermission::MANAGE_CHANNELS) {
                return Err(Error::NotAllowed);
            }
        }

        // ...
        let conn = match self.0.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = execute!(&conn, "DELETE FROM channels WHERE id = $1", &[&(id as i64)]);

        if let Err(e) = res {
            return Err(Error::DatabaseError(e.to_string()));
        }

        // delete messages
        let res = execute!(
            &conn,
            "DELETE FROM messages WHERE channel = $1",
            &[&(id as i64)]
        );

        if let Err(e) = res {
            return Err(Error::DatabaseError(e.to_string()));
        }

        // ...
        self.0.1.remove(format!("atto.channel:{}", id)).await;
        Ok(())
    }

    pub async fn add_channel_member(&self, id: usize, user: User, member: String) -> Result<()> {
        let mut y = self.get_channel_by_id(id).await?;

        if user.id != y.owner && member != user.username {
            if !user.permissions.check(FinePermission::MANAGE_CHANNELS) {
                return Err(Error::NotAllowed);
            } else {
                self.create_audit_log_entry(AuditLogEntry::new(
                    user.id,
                    format!("invoked `add_channel_member` with x value `{member}`"),
                ))
                .await?
            }
        }

        // check permissions
        let member = self.get_user_by_username(&member).await?;

        if self
            .get_userblock_by_initiator_receiver(member.id, user.id)
            .await
            .is_ok()
        {
            return Err(Error::NotAllowed);
        }

        // ...
        y.members.push(member.id);

        let conn = match self.0.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = execute!(
            &conn,
            "UPDATE channels SET members = $1 WHERE id = $2",
            params![&serde_json::to_string(&y.members).unwrap(), &(id as i64)]
        );

        if let Err(e) = res {
            return Err(Error::DatabaseError(e.to_string()));
        }

        self.0.1.remove(format!("atto.channel:{}", id)).await;

        Ok(())
    }

    pub async fn remove_channel_member(&self, id: usize, user: User, member: usize) -> Result<()> {
        let mut y = self.get_channel_by_id(id).await?;

        if user.id != y.owner && member != user.id {
            if !user.permissions.check(FinePermission::MANAGE_CHANNELS) {
                return Err(Error::NotAllowed);
            } else {
                self.create_audit_log_entry(AuditLogEntry::new(
                    user.id,
                    format!("invoked `remove_channel_member` with x value `{member}`"),
                ))
                .await?
            }
        }

        y.members
            .remove(match y.members.iter().position(|x| *x == member) {
                Some(i) => i,
                None => return Err(Error::GeneralNotFound("member".to_string())),
            });

        let conn = match self.0.connect().await {
            Ok(c) => c,
            Err(e) => return Err(Error::DatabaseConnection(e.to_string())),
        };

        let res = execute!(
            &conn,
            "UPDATE channels SET members = $1 WHERE id = $2",
            params![&serde_json::to_string(&y.members).unwrap(), &(id as i64)]
        );

        if let Err(e) = res {
            return Err(Error::DatabaseError(e.to_string()));
        }

        self.0.1.remove(format!("atto.channel:{}", id)).await;

        Ok(())
    }

    auto_method!(update_channel_title(&str)@get_channel_by_id:MANAGE_CHANNELS -> "UPDATE channels SET title = $1 WHERE id = $2" --cache-key-tmpl="atto.channel:{}");
    auto_method!(update_channel_position(i32)@get_channel_by_id:MANAGE_CHANNELS -> "UPDATE channels SET position = $1 WHERE id = $2" --cache-key-tmpl="atto.channel:{}");
    auto_method!(update_channel_minimum_role_read(i32)@get_channel_by_id:MANAGE_CHANNELS -> "UPDATE channels SET minimum_role_read = $1 WHERE id = $2" --cache-key-tmpl="atto.channel:{}");
    auto_method!(update_channel_minimum_role_write(i32)@get_channel_by_id:MANAGE_CHANNELS -> "UPDATE channels SET minimum_role_write = $1 WHERE id = $2" --cache-key-tmpl="atto.channel:{}");
    auto_method!(update_channel_members(Vec<usize>)@get_channel_by_id:MANAGE_CHANNELS -> "UPDATE channels SET members = $1 WHERE id = $2" --serde --cache-key-tmpl="atto.channel:{}");
    auto_method!(update_channel_last_message(i64) -> "UPDATE channels SET last_message = $1 WHERE id = $2" --cache-key-tmpl="atto.channel:{}");
}
